/*
 * Copyright (c) 2009 Maxim Bourmistrov <maxim@unixconn.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* This file is a part of trafscrambler
 * - an anti-sniffer/IDS NKE for OSX.
 *
 * http://en.roolz.org/trafscrambler.html
 */

#include <sys/sys_domain.h>
#include <sys/kern_control.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "common.h"

void usage(void);
void setopt(int);
const char *lookup_opt(char *, const char **);

int fd;
const char *opt;

static const char *opt_list[] = {
	"decoy", "fakedata", "zerowin",
	"cksum", "seq", "all",
	"nke", "info", NULL
};

void
setopt(int opt)
{
	if ((setsockopt(fd, SYSPROTO_CONTROL, opt, NULL, 0)) == 1) {
		fprintf(stdout, "Failed\n");
		exit(-1);
        }

}

/* taken from pfctl.c (OpenBSD) */
const char *
lookup_opt(char *cmd, const char **list)
{
        if (cmd != NULL && *cmd)
                for (; *list; list++)
                        if (!strncmp(cmd, *list, strlen(cmd)))
                                return (*list);
        return (NULL);
}

int
main (int argc, char *argv[])
{
	struct sockaddr_ctl addr;
	struct ctl_info info;

	int ch, ret;
	int enable = -1;

	while((ch = getopt(argc, argv, "d:e:")) != -1) {
		switch(ch) {
			case 'd':
				enable = 0;
				opt = lookup_opt(optarg, opt_list);
				break;
			case 'e':
				enable = 1;
				opt = lookup_opt(optarg, opt_list);
				break;
			case '?':
			default: usage();
		}
	}

	if (argc != 3)
		usage();

	fd = socket(PF_SYSTEM, SOCK_DGRAM, SYSPROTO_CONTROL);
	if (fd != -1) {
		bzero(&addr, sizeof(addr));
		bzero(&info, sizeof(info));

		strncpy(info.ctl_name, NKE_NAME, sizeof(info.ctl_name));
		if (ioctl(fd, CTLIOCGINFO, &info)) {
                	perror("Could not get ID for kernel control.\n");
                	exit(-1);
            	}
		
		addr.sc_len = sizeof(addr);
        	addr.sc_family = AF_SYSTEM;
        	addr.ss_sysaddr = AF_SYS_CONTROL;
		addr.sc_id = info.ctl_id;
            	addr.sc_unit = 0;
		
		if ((connect(fd, (struct sockaddr *)&addr, sizeof(addr))) == 1) { 
			perror("Could not connect to NKE.\n");
			exit(-1);
		}	
	}

	if (opt != NULL) {
		switch (*opt) {
		case 'd':
			if (enable)
				setopt(ENABLE_DECOY);
			else
				setopt(DISABLE_DECOY);
			break;
		case 'f':
			if (enable)
				setopt(ENABLE_DATA);
			else
				setopt(DISABLE_DATA);
			break;
		case 'a':
			if (enable)
				setopt(ENABLE_ALL);
			else
				setopt(DISABLE_ALL);
			break;
		case 'z':
			if (enable)
				setopt(ENABLE_ZERO);
			else
				setopt(DISABLE_ZERO);
			break;
		case 'c':
			if (enable)
				setopt(ENABLE_CKSUM);
			else
				setopt(DISABLE_CKSUM);
			break;
		case 's':
			if (enable)
				setopt(ENABLE_SEQ);
			else
				setopt(DISABLE_SEQ);
			break;
		case 'n':
			if (enable)
				setopt(ENABLE_NKE);
			else
				setopt(DISABLE_NKE);
			break;
		case 'i':
			if (enable)
				setopt(SHOW_INFO);
			else
				setopt(HIDE_INFO);
			break;
		}
	} else
		usage();

	(void)fprintf(stdout, "Successful\n");

	return 0;
}

void
usage(void)
{
	(void)fprintf(stderr,
		"\ntsctrl - control the trafscrambler (anti-sniffer/IDS NKE)\n"
		"[http://en.roolz.org/trafscrambler.html]\n\n"
		"usage: tsctrl [-d|-e]\n"
		"\t[-e|-d] - to enable or disable options below\n\n");
	(void)fprintf(stderr, "\tAvailable options:\n"
			"\tinfo\t\t- info regarding type of injected pkts(seen in dmesg)\n"
			"\tall\t\t- all listed below\n"
			"\tnke\t\t- operation of NKE itself\n"
			"\tdecoy\t\t- SYN-decoy pkts\n"
			"\tfakedata\t- pkts with fake data\n"
			"\tzerowin\t\t- pkts with zero win set\n"
			"\tcksum\t\t- pkts with bad TCP cksum\n"
			"\tseq\t\t- pkts with bad TCP sequenses\n");
	exit(1);
}

